import sys
import logging
import json
from subprocess import Popen, PIPE

from smac.tae.execute_ta_run import StatusType
from smac.stats.stats import Stats

__author__ = "Marius Lindauer"
__copyright__ = "Copyright 2015, ML4AAD"
__license__ = "AGPLv3"
__maintainer__ = "Marius Lindauer"
__email__ = "lindauer@cs.uni-freiburg.de"
__version__ = "0.0.1"


class ExecuteTARunAClib(object):

    """
        executes a target algorithm run with a given configuration
        on a given instance and some resource limitations
        Uses the AClib 2.0 style

        Attributes
        ----------
        ta : string
            the command line call to the target algorithm (wrapper)
        run_obj: str
            run objective (runtime or quality)
        par_factor: int
            penalized average runtime factor
    """

    def __init__(self, ta, run_obj="runtime", par_factor=1):
        """
        Constructor

        Parameters
        ----------
            ta : list
                target algorithm command line as list of arguments
            run_obj: str
                run objective of SMAC
            par_factor: int
                penalized average runtime factor
        """
        self.ta = ta
        self.logger = logging.getLogger("ExecuteTARun")
        self.run_obj = run_obj
        self.par_factor = par_factor

    def run(self, config, instance=None,
            cutoff=99999999999999.,
            seed=12345,
            instance_specific="0"):
        """
            runs target algorithm <self.ta> with configuration <config> on
            instance <instance> with instance specifics <specifics>
            for at most <cutoff> seconds and random seed <seed>

            Parameters
            ----------
                config : dictionary (or similar)
                    dictionary param -> value
                instance : string
                    problem instance
                cutoff : double
                    runtime cutoff
                seed : int
                    random seed
                instance_specific: str
                    instance specific information -- ignored here
            Returns
            -------
                status: enum of StatusType (int)
                    {SUCCESS, TIMEOUT, CRASHED, ABORT}
                cost: float
                    cost/regret/quality/runtime (float) (None, if not returned by TA)
                runtime: float
                    runtime (None if not returned by TA)
                additional_info: dict
                    all further additional run information
        """

        Stats.ta_runs += 1

        if instance is None:
            instance = "0"

        # TOOD: maybe replace fixed instance specific and cutoff_length (0) to
        # other value
        cmd = []
        cmd.extend(self.ta)
        cmd.extend(["--instance", instance,
                    "--cutoff", str(cutoff),
                    "--seed", str(seed),
                    "--config"
                    ])

        for p in config:
            if not config[p] is None:
                cmd.extend(["-" + str(p), str(config[p])])

        self.logger.debug("Calling: %s" % (" ".join(cmd)))
        p = Popen(cmd, shell=False, stdout=PIPE, stderr=PIPE, universal_newlines=True)
        stdout_, stderr_ = p.communicate()

        self.logger.debug("Stdout: %s" % (stdout_))
        self.logger.debug("Stderr: %s" % (stderr_))

        results = {"status": "CRASHED",
                   "cost": 1234567890
                   }
        for line in stdout_.split("\n"):
            if line.startswith("Result of this algorithm run:"):
                fields = ":".join(line.split(":")[1:])
                results = json.loads(fields)

        if results["status"] in ["SAT", "UNSAT", "SUCCESS"]:
            status = StatusType.SUCCESS
        elif results["status"] in ["TIMEOUT"]:
            status = StatusType.TIMEOUT
        elif results["status"] in ["CRASHED"]:
            status = StatusType.CRASHED
        elif results["status"] in ["ABORT"]:
            status = StatusType.ABORT
            self.logger.error("Target algorithm returned ABORT -- Exit!")
            sys.exit(43)
        elif results["status"] in ["MEMOUT"]:
            status = StatusType.MEMOUT

        if status in [StatusType.CRASHED, StatusType.ABORT]:
            self.logger.warn(
                "Target algorithm crashed. Last 5 lines of stdout and stderr")
            self.logger.warn("\n".join(stdout_.split("\n")[-5:]))
            self.logger.warn("\n".join(stderr_.split("\n")[-5:]))

        if results.get("runtime") is None:
            self.logger.warn(
                "The target algorithm has not returned a runtime -- imputed by 0.")
            # (TODO) Check 0
            results["runtime"] = 0

        runtime = float(results["runtime"])
        
        Stats.ta_time_used += float(runtime)

        if self.run_obj == "quality" and results.get("cost") is None:
            self.logger.error(
                "The target algorithm has not returned a quality/cost value" +
                "although we optimize cost.")
            # (TODO) Do not return 0
            results["cost"] = 0

        if self.run_obj == "runtime":
            if status != StatusType.SUCCESS:
                cost = float(results["runtime"]) * self.par_factor
            else:
                cost = float(results["runtime"])
        else:
            cost = float(results["cost"])

        del results["status"]
        try:
            del results["runtime"]
        except KeyError:
            pass
        try:
            del results["cost"]
        except KeyError:
            pass

        return status, cost, runtime, results



from abc import ABCMeta, abstractmethod
import logging

__author__ = "Katharina Eggensperger"
__copyright__ = "Copyright 2015, ML4AAD"
__license__ = "AGPLv3"
__maintainer__ = "Katharina Eggensperger"
__email__ = "eggenspk@cs.uni-freiburg.de"
__version__ = "0.0.1"


class BaseImputor(object):
    """abstract Imputor class"""

    def __init__(self):
        """
        initialize imputator module
        """
        pass

    @abstractmethod
    def impute(self, censored_x, censored_y, uncensored_x, uncensored_y):
        """
        impute runs and returns imputed y values

        Parameters
        ----------
        censored_x : array
        censored_y : list
        uncensored_x : array
        uncensored_y : list
        """


import collections
import json
import numpy

from smac.configspace import Configuration

__author__ = "Marius Lindauer"
__copyright__ = "Copyright 2015, ML4AAD"
__license__ = "AGPLv3"
__maintainer__ = "Marius Lindauer"
__email__ = "lindauer@cs.uni-freiburg.de"
__version__ = "0.0.1"

MAXINT = 2 ** 31 - 1


class RunHistory(object):

    '''
         saves all run informations from target algorithm runs

        Attributes
        ----------
    '''

    def __init__(self):
        '''
        Constructor
        '''

        # By having the data in a deterministic order we can do useful tests
        # when we serialize the data and can assume it's still in the same
        # order as it was added.
        self.data = collections.OrderedDict()

        self.RunKey = collections.namedtuple(
            'RunKey', ['config_id', 'instance_id', 'seed'])

        self.RunValue = collections.namedtuple(
            'RunValue', ['cost', 'time', 'status', 'additional_info'])

        self.config_ids = {}  # config -> id
        self.ids_config = {}  # id -> config
        self._n_id = 0

        self.cost_per_config = {} # config_id -> cost

    def add(self, config, cost, time,
            status, instance_id=None,
            seed=None,
            additional_info=None):
        '''
        adds a data of a new target algorithm (TA) run;
        it will update data if the same key values are used 
        (config, instance_id, seed)

        Attributes
        ----------
            config : dict (or other type -- depending on config space module)
                parameter configuratoin
            cost: float
                cost of TA run (will be minimized)
            time: float
                runtime of TA run
            status: str
                status in {SUCCESS, TIMEOUT, CRASHED, ABORT, MEMOUT}
            instance_id: str
                str representing an instance (default: None)
            seed: int
                random seed used by TA (default: None)
            additional_info: dict
                additional run infos (could include further returned
                information from TA or fields such as start time and host_id)
        '''

        # TODO: replace str casting of config when we have something hashable
        # as a config object
        # TODO JTS: We might have to execute one config multiple times
        #           since the results can be noisy and then we can't simply
        #           overwrite the old config result here!
        config_id = self.config_ids.get(config.__repr__())
        if config_id is None:
            self._n_id += 1
            self.config_ids[config.__repr__()] = self._n_id
            config_id = self.config_ids.get(config.__repr__())
            self.ids_config[self._n_id] = config

        k = self.RunKey(config_id, instance_id, seed)
        v = self.RunValue(cost, time, status, additional_info)

        self.data[k] = v

    def update_cost(self, config, cost):
        config_id = self.config_ids[config.__repr__()]
        self.cost_per_config[config_id] = cost

    def get_cost(self, config):
        config_id = self.config_ids[config.__repr__()]
        return self.cost_per_config[config_id]

    def get_runs_for_config(self, config):
        """Return all runs (instance seed pairs) for a configuration.

        Parameters
        ----------
        config : Configuration from ConfigSpace
            parameter configuration
        Returns
        ----------
            list: tuples of instance, seed, time
        """
        InstanceSeedPair = collections.namedtuple("InstanceSeedPair",
                                                  ["instance", "seed"])
        list_ = []
        for k in self.data:
            if config == self.ids_config[k.config_id]:
                ist = InstanceSeedPair(k.instance_id, k.seed)
                list_.append(ist)
        return list_

    def empty(self):
        """
        Check whether or not the RunHistory is empty.

        Returns
        ----------
            bool: True if runs have been added to the RunHistory, 
                  False otherwise
        """
        return len(self.data) == 0

    def save_json(self, fn="runhistory.json"):
        '''
        saves runhistory on disk

        Parameters
        ----------
        fn : str
            file name
        '''

        id_vec = dict([(id_, conf.get_array().tolist())
                       for id_, conf in self.ids_config.items()])

        data = [([int(k.config_id),
                  str(k.instance_id) if k.instance_id is not None else None,
                  int(k.seed)], list(v))
                for k, v in self.data.items()]

        with open(fn, "w") as fp:
            json.dump({"data": data,
                       "id_config": id_vec}, fp)

    def load_json(self, fn, cs):
        """Load and runhistory in json representation from disk.

        Overwrites current runthistory!

        Parameters
        ----------
        fn : str
            file name to load from
        cs : ConfigSpace
            instance of configuration space
        """

        with open(fn) as fp:
            all_data = json.load(fp)

        self.ids_config = dict([(int(id_), Configuration(
            cs, vector=numpy.array(vec))) for id_, vec in all_data["id_config"].items()])


        self.config_ids = dict([(Configuration(
            cs, vector=numpy.array(vec)).__repr__(), id_) for id_, vec in all_data["id_config"].items()])
        self._n_id = len(self.config_ids)
        
        self.data = dict([(self.RunKey(int(k[0]), k[1], int(k[2])),
                           self.RunValue(float(v[0]), float(v[1]), v[2], v[3]))
                          for k, v in all_data["data"]
                          ])

    def update_from_json(self, fn, cs):
        """Update the current runhistory by adding new runs from a json file.

        Parameters
        ----------
        fn : str
            file name to load from
        cs : ConfigSpace
            instance of configuration space
        """

        new_runhistory = RunHistory()
        new_runhistory.load_json(fn, cs)

        # Configurations might be already known, but by a different ID. This
        # does not matter here because the add() method handles this
        # correctly by assigning an ID to unknown configurations and re-using
        #  the ID
        for key, value in new_runhistory.data.items():
            config_id, instance_id, seed = key
            cost, time, status, additional_info = value
            config = new_runhistory.ids_config[config_id]
            self.add(config=config, cost=cost, time=time,
                     status=status, instance_id=instance_id,
                     seed=seed, additional_info=additional_info)



__author__ = "Marius Lindauer"
__copyright__ = "Copyright 2016, ML4AAD"
__license__ = "AGPLv3"
__maintainer__ = "Marius Lindauer"
__email__ = "lindauer@cs.uni-freiburg.de"
__version__ = "0.0.1"

import os
import logging
import numpy
import json

from smac.stats.stats import Stats


class TrajLogger(object):

    """
        writes trajectory logs files 

        Attributes
        ----------
        logger : Logger oject
    """

    def __init__(self, output_dir):
        """
        Constructor 

        creates output directory if not exists already

        Arguments
        ---------
        output_dir: str
            directory for logging
        """
        self.logger = logging.getLogger("TrajLogger")

        if not os.path.isdir(output_dir):
            try:
                os.makedirs(output_dir)
            except OSError:
                self.logger.error(
                    "Could not make output directory: %s" % (output_dir))

        self.output_dir = output_dir

        self.old_traj_fn = os.path.join(output_dir, "traj_old.csv")
        if not os.path.isfile(self.old_traj_fn):
            with open(self.old_traj_fn, "w") as fp:
                fp.write(
                    '"CPU Time Used","Estimated Training Performance","Wallclock Time","Incumbent ID","Automatic Configurator (CPU) Time","Configuration..."\n')

        self.aclib_traj_fn = os.path.join(output_dir, "traj_aclib2.json")

    def add_entry(self, train_perf,
                  incumbent_id, incumbent):
        """
            adds entries to trajectory files (several formats)

            Parameters
            ----------
            train_perf: float
                estimated performance on training (sub)set 
            incumbent_id: int
                id of incumbent
            incumbent: Configuration()
                current incumbent configuration
        """

        self._add_in_old_format(train_perf, incumbent_id, incumbent)
        self._add_in_aclib_format(train_perf, incumbent_id, incumbent)

    def _add_in_old_format(self, train_perf,
                           incumbent_id, incumbent):
        """
            adds entries to old SMAC2-like trajectory file

            Parameters
            ----------
            train_perf: float
                estimated performance on training (sub)set 
            incumbent_id: int
                id of incumbent
            incumbent: Configuration()
                current incumbent configuration
        """

        conf = []
        for p in incumbent:
            if not incumbent[p] is None:
                conf.append("%s='%s'" % (p, incumbent[p]))

        ta_time_used = Stats.ta_time_used
        wallclock_time = Stats.get_used_wallclock_time()

        with open(self.old_traj_fn, "a") as fp:
            fp.write("%f, %f, %f, %d, %f, %s\n" % (
                ta_time_used,
                train_perf,
                wallclock_time,
                incumbent_id,
                wallclock_time - ta_time_used,
                ", ".join(conf)
            ))

    def _add_in_aclib_format(self, train_perf,
                             incumbent_id, incumbent):
        """
            adds entries to AClib2-like trajectory file

            Parameters
            ----------
            train_perf: float
                estimated performance on training (sub)set 
            incumbent_id: int
                id of incumbent
            incumbent: Configuration()
                current incumbent configuration
        """

        conf = []
        for p in incumbent:
            if not incumbent[p] is None:
                conf.append("%s='%s'" % (p, incumbent[p]))

        ta_time_used = Stats.ta_time_used
        wallclock_time = Stats.get_used_wallclock_time()

        traj_entry = {"cpu_time": ta_time_used,
                      "total_cpu_time": None,  # TODO: fix this
                      "wallclock_time": wallclock_time,
                      "evaluations": Stats.ta_runs,
                      "cost": train_perf,
                      "incumbent": conf
                      }

        with open(self.aclib_traj_fn, "a") as fp:
            json.dump(traj_entry, fp)
            fp.write("\n")



import copy
from collections import OrderedDict
import logging
import sys
import time
import random
from collections import Counter

import numpy

from smac.tae.execute_ta_run_aclib import ExecuteTARunAClib

from smac.smbo.objective import total_runtime, sum_cost
from smac.stats.stats import Stats

__author__ = "Katharina Eggensperger, Marius Lindauer"
__copyright__ = "Copyright 2015, ML4AAD"
__license__ = "AGPLv3"
__maintainer__ = "Katharina Eggensperger"
__email__ = "eggenspk@cs.uni-freiburg.de"
__version__ = "0.0.1"

MAXINT = 2 ** 31 - 1


class Intensifier(object):

    '''
        takes challenger and incumbents and performs intensify
    '''

    def __init__(self, executor, instances=None,
                 instance_specifics={},
                 cutoff=MAXINT, deterministic=False, run_obj_time=True,
                 run_limit=MAXINT, maxR=2000, rng=0):
        '''
        Constructor

        Parameters
        ----------
        executor : tae.executre_ta_run_*.ExecuteTARun* Object
            target algorithm run executor
        instances : list
            list of all instance ids
        instance_specifics : dict
            mapping from instance name to instance specific string
        cutoff : int
            runtime cutoff of TA runs
        deterministic: bool
            whether the TA is deterministic or not
        run_obj_time: bool
            whether the run objective is runtime or not (if true, apply adaptive capping)
        run_limit : int
            maximum number of runs
        maxR : int
            maximum number of runs per config
        '''
        # general attributes
        if instances is None:
            instances = []
        self.instances = set(instances)
        self.instance_specifics = instance_specifics
        self.logger = logging.getLogger("intensifier")
        self.run_limit = run_limit
        self.maxR = maxR
        self.rs = numpy.random.RandomState(rng)

        # scenario info
        self.cutoff = cutoff
        self.deterministic = deterministic
        self.run_obj_time = run_obj_time
        self.tae = executor

        self.Adaptive_Capping_Slackfactor = 1.2

        if self.run_limit < 1:
            raise ValueError("run_limit must be > 1")

    def intensify(self, challengers, incumbent, run_history, objective,
                  time_bound=MAXINT):
        '''
            running intensification to determine the incumbent configuration
            Side effect: adds runs to run_history

            Implementation of Procedure 2 in Hutter et al. (2011).

            Parameters
            ----------

            challengers : list of ConfigSpace.config
                promising configurations
            incumbent : ConfigSpace.config
                best configuration so far
            run_history : runhistory
                all runs on all instance,seed pairs for incumbent
            time_bound : int, optional (default=2 ** 31 - 1)
                time in [sec] available to perform intensify

            Returns
            -------
            incumbent: Configuration()
                current (maybe new) incumbent configuration
            inc_perf: float
                empirical performance of incumbent configuration 
        '''

        self.start_time = time.time()

        if time_bound < 0.01:
            raise ValueError("time_bound must be >= 0.01")

        num_run = 0

        # Line 1 + 2
        for challenger in challengers:
            self.logger.debug("Intensify on %s", challenger)
            if hasattr(challenger, 'origin'):
                self.logger.debug(
                    "Configuration origin: %s", challenger.origin)
            inc_runs = run_history.get_runs_for_config(incumbent)

            # Line 3
            # First evaluate incumbent on a new instance
            if len(inc_runs) < self.maxR:
                # Line 4
                # find all instances that have the most runs on the inc
                inc_inst = [s.instance for s in inc_runs]
                inc_inst = list(Counter(inc_inst).items())
                inc_inst.sort(key=lambda x: x[1], reverse=True)
                max_runs = inc_inst[0][1]
                inc_inst = set([x[0] for x in inc_inst if x[1] == max_runs])

                available_insts = (self.instances - inc_inst)

                # if all instances were used n times, we can pick an instances
                # from the complete set again
                if not self.deterministic and not available_insts:
                    available_insts = self.instances

                # Line 6 (Line 5 is further down...)
                if self.deterministic:
                    next_seed = 0
                else:
                    next_seed = self.rs.randint(low=0, high=MAXINT,
                                                size=1)[0]

                if available_insts:
                    # Line 5 (here for easier code)
                    next_instance = random.choice(list(available_insts))
                    # Line 7
                    status, cost, dur, res = self.tae.run(config=incumbent,
                                                          instance=next_instance,
                                                          seed=next_seed,
                                                          cutoff=self.cutoff,
                                                          instance_specific=self.instance_specifics.get(next_instance, "0"))
                    run_history.add(config=incumbent,
                                    cost=cost, time=dur, status=status,
                                    instance_id=next_instance, seed=next_seed,
                                    additional_info=res)
                    # TODO update incumbent performance here!

                    num_run += 1
                else:
                    self.logger.debug(
                        "No further instance-seed pairs for incumbent available.")

            # Line 8
            N = 1

            inc_inst_seeds = set(run_history.get_runs_for_config(incumbent))

            # Line 9
            while True:
                chall_inst_seeds = set(map(lambda x: (
                    x.instance, x.seed), run_history.get_runs_for_config(challenger)))

                # Line 10
                missing_runs = list(inc_inst_seeds - chall_inst_seeds)

                # Line 11
                self.rs.shuffle(missing_runs)
                to_run = missing_runs[:min(N, len(missing_runs))]
                # Line 13 (Line 12 comes below...)
                missing_runs = missing_runs[min(N, len(missing_runs)):]

                inst_seed_pairs = list(inc_inst_seeds - set(missing_runs))
                inc_perf = objective(incumbent, run_history, inst_seed_pairs)
                run_history.update_cost(incumbent, inc_perf)

                inc_sum_cost = sum_cost(config=incumbent, instance_seed_pairs=inst_seed_pairs,
                                        run_history=run_history)

                # Line 12
                for instance, seed in to_run:
                    # Run challenger on all <config,seed> to run
                    if self.run_obj_time:
                        chall_inst_seeds = set(map(lambda x: (
                            x.instance, x.seed), run_history.get_runs_for_config(challenger)))
                        chal_sum_cost = sum_cost(config=challenger, instance_seed_pairs=chall_inst_seeds,
                                                 run_history=run_history)
                        cutoff = min(self.cutoff,
                                     (inc_sum_cost - chal_sum_cost) *
                                     self.Adaptive_Capping_Slackfactor)

                        if cutoff < 0:  # no time left to validate challenger
                            self.logger.debug("Stop challenger itensification due to adaptive capping.")
                            break

                    else:
                        cutoff = self.cutoff

                    status, cost, dur, res = self.tae.run(config=challenger,
                                                          instance=instance,
                                                          seed=seed,
                                                          cutoff=cutoff,
                                                          instance_specific=self.instance_specifics.get(instance, "0"))

                    run_history.add(config=challenger,
                                    cost=cost, time=dur, status=status,
                                    instance_id=instance, seed=seed,
                                    additional_info=res)
                    num_run += 1

                # we cannot use inst_seed_pairs here since we could have less runs
                # for the challenger due to the adaptive capping
                chall_inst_seeds = set(map(lambda x: (
                    x.instance, x.seed), run_history.get_runs_for_config(challenger)))
                chal_perf = objective(challenger, run_history, chall_inst_seeds)
                run_history.update_cost(challenger, chal_perf)

                # Line 15
                if chal_perf > inc_perf:
                    # Incumbent beats challenger
                    self.logger.debug("Incumbent (%.4f) is better than challenger (%.4f) on %d runs." % (
                        inc_perf, chal_perf, len(inst_seed_pairs)))
                    break
                # Line 16
                elif len(missing_runs) == 0:
                    # Challenger is as good as incumbent -> change incu

                    n_samples = len(inst_seed_pairs)
                    self.logger.info("Challenger (%.4f) is better than incumbent (%.4f) on %d runs." % (
                        chal_perf / n_samples, inc_perf / n_samples, n_samples))
                    self.logger.info(
                        "Changing incumbent to challenger: %s" % (challenger))
                    incumbent = challenger
                    inc_perf = chal_perf
                    Stats.inc_changed += 1
                    break
                # Line 17
                else:
                    # challenger is not worse, continue
                    N = 2 * N

            if num_run > self.run_limit:
                self.logger.debug(
                    "Maximum #runs for intensification reached")
                break
            elif time.time() - self.start_time - time_bound >= 0:
                self.logger.debug("Timelimit for intensification reached ("
                                  "used: %f sec, available: %f sec)" % (
                                      time.time() - self.start_time, time_bound))
                break

        # output estimated performance of incumbent
        inc_runs = run_history.get_runs_for_config(incumbent)
        inc_perf = objective(incumbent, run_history, inc_runs)
        self.logger.info("Updated estimated performance of incumbent on %d runs: %.4f" % (
            len(inc_runs), inc_perf))

        return incumbent, inc_perf



__author__ = "Marius Lindauer"
__copyright__ = "Copyright 2015, ML4AAD"
__license__ = "AGPLv3"
__maintainer__ = "Marius Lindauer"
__email__ = "lindauer@cs.uni-freiburg.de"
__version__ = "0.0.1"

import numpy as np
from smac.configspace import pcs

class InputReader(object):

    """
        reading all input files for SMAC (scenario file, instance files, ...)

        Note: most of this code was taken from the pysmac repository.
              We copy it here because we don't want smac3 to depend
              on an earlier version!
    """

    def __init__(self):
        """
        Constructor
        """
        pass

    def read_scenario_file(self, fn):
        """
            encapsulates read_scenario_file of pysmac

            Returns
            -------
            dict : dictionary
                (key, value) pairs are (variable name, variable value)
        """
        # translate the difference option names to a canonical name
        scenario_option_names = {'algo-exec' : 'algo', 
                            'algoExec': 'algo',
                            'algo-exec-dir': 'execdir',
                            'exec-dir' : 'execdir',
                            'execDir' : 'execdir',
                            'algo-deterministic' : 'deterministic',
                            'paramFile' : 'paramfile',
                            'pcs-file' :'paramfile',
                            'param-file' : 'paramfile',
                            'run-obj' : 'run_obj',
                            'run-objective' : 'run_obj',
                            'runObj' : 'run_obj',
                            'overall_obj' : 'overall_obj',
                            'intra-obj' : 'overall_obj', 
                            'intra-instance-obj'  : 'overall_obj', 
                            'overall-obj'  : 'overall_obj', 
                            'intraInstanceObj' : 'overall_obj',
                            'overallObj' : 'overall_obj', 
                            'intra_instance_obj' : 'overall_obj',
                            'algo-cutoff-time' : 'cutoff_time', 
                            'target-run-cputime-limit' : 'cutoff_time',
                            'target_run_cputime_limit'  : 'cutoff_time', 
                            'cutoff-time' : 'cutoff_time', 
                            'cutoffTime' : 'cutoff_time',
                            'cputime-limit' : 'tunerTimeout', 
                            'cputime_limit' : 'tunerTimeout', 
                            'tunertime-limit' : 'tunerTimeout', 
                            'tuner-timeout' : 'tunerTimeout', 
                            'tunerTimeout' : 'tunerTimeout',
                            'wallclock_limit' : 'wallclock-limit',  
                            'runtime-limit' : 'wallclock-limit', 
                            'runtimeLimit' : 'wallclock-limit', 
                            'wallClockLimit'  : 'wallclock-limit',
                            'output-dir' : 'outdir', 
                            'outputDirectory' : 'outdir',
                            'instances' : 'instance_file', 
                            'instance-file' : 'instance_file' , 
                            'instance-dir' : 'instance_file', 
                            'instanceFile' : 'instance_file', 
                            'i' : 'instance_file', 
                            'instance_seed_file' : 'instance_file',
                            'test-instances' : 'test_instance_file', 
                            'test-instance-file'  : 'test_instance_file', 
                            'test-instance-dir' : 'test_instance_file', 
                            'testInstanceFile' : 'test_instance_file', 
                            'test_instance_file' : 'test_instance_file', 
                            'test_instance_seed_file' : 'test_instance_file',               
                            'feature-file' : 'feature_file', 
                            'instanceFeatureFile' : 'feature_file', 
                            'feature_file' : 'feature_file',
                            'runcount-limit' : 'runcount-limit',
                            'runcount_limit' : 'runcount-limit',
                            'totalNumRunsLimit' : 'runcount-limit',
                            'numRunsLimit' : 'runcount-limit',
                            'numberOfRunsLimit' : 'runcount-limit'
                        }

        scenario_dict = {}
        with open(fn, 'r') as fh:
            for line in fh.readlines():
                #remove comments
                if line.find("#") > -1:
                    line = line[:line.find("#")]

                # skip empty lines
                if line  == "":
                    continue
                if "=" in line:
                    tmp = line.split("=")
                    tmp = [' '.join(s.split()) for s in tmp]
                else:
                    tmp = line.split()
                scenario_dict[scenario_option_names.get(tmp[0],tmp[0])] = " ".join(tmp[1:])
        return(scenario_dict)

    def read_instance_file(self, fn):
        """
            encapsulates read_instances_file of pysmac

            Parameters
            ----------
                fn: string
                     file name of instance file

            Returns
            -------
                list -- each element is a list where the first element is the
                instance name followed by additional
                information for the specific instance.
        """
        with open(fn,'r') as fh:
            instance_names = fh.readlines()
        return([s.strip().split() for s in instance_names])

    def read_instance_features_file(self, fn):
        """
            encapsulates read_instances_file of pysmac

            Parameters
            ----------
                fn: string
                     file name of instance feature file

            Returns
            -------
                tuple -- first entry is a list of the feature names,
                second one is a dict with 'instance name' -
                'numpy array containing the features' key-value pairs
        """
        instances = {}
        with open(fn,'r') as fh:
            lines = fh.readlines()
            for line in lines[1:]:
                tmp = line.strip().split(",")
                instances[tmp[0]] = np.array(tmp[1:],dtype=np.double)
        return(lines[0].split(",")[1:], instances)

    def read_pcs_file(self, fn):
        """
            encapsulates generating configuration space object

            Parameters
            ----------
                fn: string
                     file name of pcs file

            Returns
            -------
                Object of ConfigSpace
        """
        space = pcs.read(fn)
        return space



import time
import logging

__author__ = "Marius Lindauer"
__copyright__ = "Copyright 2016, ML4AAD"
__license__ = "AGPLv3"
__maintainer__ = "Marius Lindauer"
__email__ = "lindauer@cs.uni-freiburg.de"
__version__ = "0.0.1"


class Stats(object):

    '''
        all statistics collected during configuration run
        This is a static class without initialization to get an easy access every in the code
    '''
    
    scenario = None
    
    ta_runs = 0
    wallclock_time_used = 0
    incumbent_changed = 0
    ta_time_used = 0
    inc_changed = 0

    _start_time = None
    _logger = logging.getLogger("Stats")

    @staticmethod
    def start_timing():
        '''
            starts the timer (for the runtime configuration budget)
        '''
        if Stats.scenario:
            Stats._start_time = time.time()
        else:
            raise ValueError("Scenario is missing")

    @staticmethod
    def get_used_wallclock_time():
        '''
            returns used wallclock time
            Returns
            -------
            wallclock_time : int
                used wallclock time in sec
        '''
        
        return time.time() - Stats._start_time

    @staticmethod
    def get_remaing_time_budget():
        '''
            subtracts the runtime configuration budget with the used wallclock time
        '''
        if Stats.scenario:
            return Stats.scenario.wallclock_limit - (time.time() - Stats._start_time)
        else:
            raise "Scenario is missing"

    @staticmethod
    def get_remaining_ta_runs():
        '''
           subtract the target algorithm runs in the scenario with the used ta runs 
        '''
        if Stats.scenario:
            return Stats.scenario.ta_run_limit - Stats.ta_runs
        else:
            raise "Scenario is missing"

    @staticmethod
    def get_remaining_ta_budget():
        '''
            subtracts the ta running budget with the used time
        '''
        if Stats.scenario:
            return Stats.scenario.algo_runs_timelimit - Stats.ta_time_used

    @staticmethod
    def print_stats():
        '''
            prints all statistics
        '''
        Stats._logger.info("##########################################################")
        Stats._logger.info("Statistics:")
        Stats._logger.info("#Target algorithm runs: %d" %(Stats.ta_runs))
        Stats._logger.info("Used wallclock time: %.2f sec" %(time.time() - Stats._start_time))
        Stats._logger.info("Used target algorithm runtime: %.2f sec" %(Stats.ta_time_used))
        
        Stats._logger.info("##########################################################")    



import itertools
import logging
import numpy as np
import os
import random
import sys
import time

from ConfigSpace.hyperparameters import CategoricalHyperparameter, \
    UniformFloatHyperparameter, UniformIntegerHyperparameter, Constant
import ConfigSpace.util

from smac.smbo.acquisition import EI
from smac.smbo.base_solver import BaseSolver
from smac.epm.rf_with_instances import RandomForestWithInstances
from smac.smbo.local_search import LocalSearch
from smac.smbo.intensification import Intensifier
from smac.smbo import pSMAC
from smac.runhistory.runhistory import RunHistory
from smac.runhistory.runhistory2epm import RunHistory2EPM4Cost, \
    RunHistory2EPM4LogCost
from smac.smbo.objective import average_cost, total_runtime
from smac.tae.execute_ta_run import StatusType
from smac.stats.stats import Stats
from smac.tae.execute_ta_run_old import ExecuteTARunOld
from smac.utils.io.traj_logging import TrajLogger

from smac.epm.rfr_imputator import RFRImputator

MAXINT = 2 ** 31 - 1

__author__ = "Aaron Klein, Marius Lindauer, Matthias Feurer"
__copyright__ = "Copyright 2015, ML4AAD"
__license__ = "AGPLv3"
#__maintainer__ = "???"
#__email__ = "???"
__version__ = "0.0.1"


def get_types(config_space, instance_features=None):
    # Extract types vector for rf from config space
    types = np.zeros(len(config_space.get_hyperparameters()),
                     dtype=np.uint)

    for i, param in enumerate(config_space.get_hyperparameters()):
        if isinstance(param, (CategoricalHyperparameter)):
            n_cats = len(param.choices)
            types[i] = n_cats

        elif isinstance(param, Constant):
            # for constants we simply set types to 0
            # which makes it a numerical parameter
            types[i] = 0
            # and we leave the bounds to be 0 for now
        elif not isinstance(param, (UniformFloatHyperparameter,
                                    UniformIntegerHyperparameter)):
            raise TypeError("Unknown hyperparameter type %s" % type(param))

    if instance_features is not None:
        types = np.hstack(
            (types, np.zeros((instance_features.shape[1]))))

    types = np.array(types, dtype=np.uint)
    return types


class SMBO(BaseSolver):

    def __init__(self, scenario, tae_runner=None, acquisition_function=None,
                 model=None, runhistory2epm=None, rng=None):
        '''
        Interface that contains the main Bayesian optimization loop

        Parameters
        ----------
        scenario: smac.scenario.scenario.Scenario
            Scenario object
        tae_runner: object
            object that implements the following method to call the target
            algorithm (or any other arbitrary function):
            run(self, config)
            If not set, it will be initialized with the tae.ExecuteTARunOld()
        acquisition_function : AcquisitionFunction
            Object that implements the AbstractAcquisitionFunction. Will use
            EI if not set.
        model : object
            Model that implements train() and predict(). Will use a
            RandomForest if not set.
        runhistory2epm : RunHistory2EMP
            Object that implements the AbstractRunHistory2EPM. If None,
            will use RunHistory2EPM4Cost if objective is cost or
            RunHistory2EPM4LogCost if objective is runtime.
        rng: numpy.random.RandomState
            Random number generator
        '''
        
        self.runhistory = RunHistory()
        
        self.logger = logging.getLogger("smbo")

        if rng is None:
            self.num_run = np.random.randint(1234567980)
            self.rng = np.random.RandomState(seed=self.num_run)
        elif isinstance(rng, int):
            self.num_run = rng
            self.rng = np.random.RandomState(seed=rng)
        elif isinstance(rng, np.random.RandomState):
            self.num_run = rng.randint(1234567980)
            self.rng = rng
        else:
            raise TypeError('Unknown type %s for argument rng. Only accepts '
                            'None, int or np.random.RandomState' % str(type(rng)))

        self.scenario = scenario
        self.config_space = scenario.cs

        self.types = get_types(self.config_space, scenario.feature_array)
        if model is None:
            self.model = RandomForestWithInstances(self.types,
                                                   scenario.feature_array,
                                                   seed=self.rng.randint(
                                                       1234567980))
        else:
            self.model = model

        if acquisition_function is None:
            self.acquisition_func = EI(self.model)
        else:
            self.acquisition_func = acquisition_function

        self.local_search = LocalSearch(self.acquisition_func,
                                        self.config_space)
        self.incumbent = None

        if tae_runner is None:
            self.executor = ExecuteTARunOld(ta=scenario.ta,
                                            run_obj=scenario.run_obj,
                                            par_factor=scenario.par_factor)
        else:
            self.executor = tae_runner

        self.inten = Intensifier(executor=self.executor,
                                 instances=self.scenario.train_insts,
                                 cutoff=self.scenario.cutoff,
                                 deterministic=self.scenario.deterministic,
                                 run_obj_time=self.scenario.run_obj == "runtime",
                                 instance_specifics=self.scenario.instance_specific)

        num_params = len(self.config_space.get_hyperparameters())

        self.objective = average_cost
        if self.scenario.run_obj == "runtime":

            if runhistory2epm is None:
                # if we log the performance data,
                # the RFRImputator will already get
                # log transform data from the runhistory
                cutoff = np.log10(self.scenario.cutoff)
                threshold = np.log10(self.scenario.cutoff *
                                     self.scenario.par_factor)

                imputor = RFRImputator(cs=self.config_space,
                                       rs=self.rng,
                                       cutoff=cutoff,
                                       threshold=threshold,
                                       model=self.model,
                                       change_threshold=0.01,
                                       max_iter=10)
                self.rh2EPM = RunHistory2EPM4LogCost(
                    scenario=self.scenario, num_params=num_params,
                    success_states=[StatusType.SUCCESS, ],
                    impute_censored_data=True,
                    impute_state=[StatusType.TIMEOUT, ],
                    imputor=imputor)
            else:
                self.rh2EPM = runhistory2epm

        elif self.scenario.run_obj == 'quality':
            if runhistory2epm is None:
                self.rh2EPM = RunHistory2EPM4Cost\
                    (scenario=self.scenario, num_params=num_params,
                     success_states=[StatusType.SUCCESS, ],
                     impute_censored_data=False, impute_state=None)
            else:
                self.rh2EPM = runhistory2epm

        else:
            raise ValueError('Unknown run objective: %s. Should be either '
                             'quality or runtime.' % self.scenario.run_obj)

        self.trajLogger = TrajLogger(self.scenario.output_dir)

    def run_initial_design(self):
        '''
            runs algorithm runs for a initial design;
            default implementation: running the default configuration on
                                    a random instance-seed pair
            Side effect: adds runs to self.runhistory

            Returns
            -------
            incumbent: Configuration()
                initial incumbent configuration
        '''

        default_conf = self.config_space.get_default_configuration()
        self.incumbent = default_conf
        rand_inst_id = self.rng.randint(0, len(self.scenario.train_insts))
        # ignore instance specific values
        rand_inst = self.scenario.train_insts[rand_inst_id]

        if self.scenario.deterministic:
            initial_seed = 0
        else:
            initial_seed = random.randint(0, MAXINT)

        status, cost, runtime, additional_info = self.executor.run(
            default_conf, instance=rand_inst, cutoff=self.scenario.cutoff,
            seed=initial_seed,
            instance_specific=self.scenario.instance_specific.get(rand_inst, "0"))

        if status in [StatusType.CRASHED or StatusType.ABORT]:
            self.logger.info("First run crashed -- Abort")
            sys.exit(42)

        self.runhistory.add(config=default_conf, cost=cost, time=runtime,
                            status=status,
                            instance_id=rand_inst,
                            seed=initial_seed,
                            additional_info=additional_info)
        defaul_inst_seeds = set(
            self.runhistory.get_runs_for_config(default_conf))
        default_perf = self.objective(default_conf, self.runhistory,
                                      defaul_inst_seeds)
        self.runhistory.update_cost(default_conf, default_perf)

        Stats.inc_changed += 1  # first incumbent
        return default_conf

    def run(self, max_iters=10):
        '''
        Runs the Bayesian optimization loop for max_iters iterations

        Parameters
        ----------
        max_iters: int
            The maximum number of iterations

        Returns
        ----------
        incumbent: np.array(1, H)
            The best found configuration
        '''
        Stats.start_timing()

        #self.runhistory = RunHisory()

        self.incumbent = self.run_initial_design()

        self.trajLogger.add_entry(train_perf=999999999,
                                  incumbent_id=Stats.inc_changed,
                                  incumbent=self.incumbent)

        inc_id = Stats.inc_changed  # ID of incumbent

        # Main BO loop
        iteration = 1
        while True:
            if self.scenario.shared_model:
                pSMAC.read(run_history=self.runhistory,
                           output_directory=self.scenario.output_dir,
                           configuration_space=self.config_space,
                           logger=self.logger)

            start_time = time.time()
            X, Y = self.rh2EPM.transform(self.runhistory)

            self.logger.debug("Search for next configuration")
            # get all found configurations sorted according to acq
            challengers = self.choose_next(X, Y)

            time_spend = time.time() - start_time
            logging.debug(
                "Time spend to choose next configurations: %.2f sec" % (time_spend))

            self.logger.debug("Intensify")

            self.incumbent, inc_perf = self.inten.intensify(
                challengers=challengers,
                incumbent=self.incumbent,
                run_history=self.runhistory,
                objective=self.objective,
                time_bound=max(0.01, time_spend))

            # TODO: Write run history into database
            if self.scenario.shared_model:
                pSMAC.write(run_history=self.runhistory,
                            output_directory=self.scenario.output_dir,
                            num_run=self.num_run)

            if iteration == max_iters:
                break

            iteration += 1

            logging.debug("Remaining budget: %f (wallclock), %f (ta costs), %f (target runs)" % (
                Stats.get_remaing_time_budget(),
                Stats.get_remaining_ta_budget(),
                Stats.get_remaining_ta_runs()))

            if Stats.get_remaing_time_budget() < 0 or \
                    Stats.get_remaining_ta_budget() < 0 or \
                    Stats.get_remaining_ta_runs() < 0:
                break

            if Stats.inc_changed > inc_id:
                self.trajLogger.add_entry(train_perf=inc_perf,
                                          incumbent_id=Stats.inc_changed,
                                          incumbent=self.incumbent)
                inc_id = Stats.inc_changed

        return self.incumbent

    def choose_next(self, X, Y, num_interleaved_random=1010,
                    num_configurations_by_random_search_sorted=1000,
                    num_configurations_by_local_search=10):
        """Choose next candidate solution with Bayesian optimization.

        Parameters
        ----------
        X : (N, D) numpy array
            Each row contains a configuration and one set of
            instance features.
        Y : (N, O) numpy array
            The function values for each configuration instance pair.

        Returns
        -------
        list
            List of 2020 suggested configurations to evaluate.
        """
        self.model.train(X, Y)

        if self.runhistory.empty():
            incumbent_value = 0.0
        elif self.incumbent is None:
            # TODO try to calculate an incumbent from the runhistory!
            incumbent_value = 0.0
        else:
            incumbent_value = self.runhistory.get_cost(self.incumbent)

        self.acquisition_func.update(model=self.model, eta=incumbent_value)

        # Remove dummy acquisition function value
        next_configs_by_random_search = [x[1] for x in
                                         self._get_next_by_random_search(num_points=num_interleaved_random)]

        # Get configurations sorted by EI
        next_configs_by_random_search_sorted = \
            self._get_next_by_random_search(
                num_configurations_by_random_search_sorted, _sorted=True)
        next_configs_by_local_search = \
            self._get_next_by_local_search(num_configurations_by_local_search)

        next_configs_by_acq_value = next_configs_by_random_search_sorted + \
            next_configs_by_local_search
        next_configs_by_acq_value.sort(reverse=True, key=lambda x: x[0])
        next_configs_by_acq_value = [_[1] for _ in next_configs_by_acq_value]

        challengers = list(itertools.chain(*zip(next_configs_by_acq_value,
                                                next_configs_by_random_search)))
        return challengers

    def _get_next_by_random_search(self, num_points=1000, _sorted=False):
        """Get candidate solutions via local search.

        Parameters
        ----------
        num_points : int, optional (default=10)
            Number of local searches and returned values.

        _sorted : bool, optional (default=True)
            Whether to sort the candidate solutions by acquisition function
            value.

        Returns
        -------
        list : (acquisition value, Candidate solutions)
        """

        rand_configs = self.config_space.sample_configuration(size=num_points)
        if _sorted:
            imputed_rand_configs = map(ConfigSpace.util.impute_inactive_values,
                                       rand_configs)
            imputed_rand_configs = [x.get_array()
                                    for x in imputed_rand_configs]
            imputed_rand_configs = np.array(imputed_rand_configs,
                                            dtype=np.float64)
            acq_values = self.acquisition_func(imputed_rand_configs)
            # From here
            # http://stackoverflow.com/questions/20197990/how-to-make-argsort-result-to-be-random-between-equal-values
            random = self.rng.rand(len(acq_values))
            # Last column is primary sort key!
            indices = np.lexsort((random.flatten(), acq_values.flatten()))

            for i in range(len(rand_configs)):
                rand_configs[i].origin = 'Random Search (sorted)'

            # Cannot use zip here because the indices array cannot index the
            # rand_configs list, because the second is a pure python list
            return [(acq_values[ind], rand_configs[ind])
                    for ind in indices[::-1]]
        else:
            for i in range(len(rand_configs)):
                rand_configs[i].origin = 'Random Search'
            return [(0, rand_configs[i]) for i in range(len(rand_configs))]

    def _get_next_by_local_search(self, num_points=10):
        """Get candidate solutions via local search.

        In case acquisition function values tie, these will be broken randomly.

        Parameters
        ----------
        num_points : int, optional (default=10)
            Number of local searches and returned values.

        Returns
        -------
        list : (acquisition value, Candidate solutions),
               ordered by their acquisition function value
        """
        configs_acq = []

        # Start N local search from different random start points
        for i in range(num_points):
            if i == 0 and self.incumbent is not None:
                start_point = self.incumbent
            else:
                start_point = self.config_space.sample_configuration()

            configuration, acq_val = self.local_search.maximize(start_point)

            configuration.origin = 'Local Search'
            configs_acq.append((acq_val[0][0], configuration))

        # shuffle for random tie-break
        random.shuffle(configs_acq, self.rng.rand)

        # sort according to acq value
        # and return n best configurations
        configs_acq.sort(reverse=True, key=lambda x: x[0])

        return configs_acq



import logging
import numpy
from scipy.stats import truncnorm

import smac.epm.base_imputor

from ConfigSpace.hyperparameters import UniformIntegerHyperparameter, \
    CategoricalHyperparameter, UniformFloatHyperparameter

__author__ = "Katharina Eggensperger"
__copyright__ = "Copyright 2015, ML4AAD"
__license__ = "AGPLv3"
__maintainer__ = "Katharina Eggensperger"
__email__ = "eggenspk@cs.uni-freiburg.de"
__version__ = "0.0.1"


class RFRImputator(smac.epm.base_imputor.BaseImputor):

    """Uses an rfr to do imputation"""

    def __init__(self, cs, rs, cutoff, threshold,
                 model,
                 change_threshold=0.01,
                 max_iter=10):
        """
        initialize imputator module

        Parameters
        ----------
        max_iter : maximum number of iteration
        cs : config space object
        rs : random state generator
        cutoff : float
            cutoff value used for this scenario
        threshold : float
            highest possible values (e.g. cutoff * par)
        model:
            epm model (i.e. RandomForestWithInstances)
        change_threshold : float 
            stop imputation if change is less than this
        -------
        """

        super(RFRImputator, self).__init__()
        self.logger = logging.getLogger("RFRImputor")
        self.max_iter = max_iter
        self.change_threshold = change_threshold
        self.cutoff = cutoff
        self.threshold = threshold
        self.seed = rs.random_integers(low=0, high=1000)

        self.model = model

        # Never use a lower variance than this
        self.var_threshold = 10 ** -2

    def impute(self, censored_X, censored_y, uncensored_X, uncensored_y):
        """
        impute runs and returns imputed y values

        Parameters
        ----------
        censored_X : array
            X matrix of censored data
        censored_y : array
            y matrix of censored data
        uncensored_X : array
            X matrix of uncensored data
        uncensored_y : array
            y matrix of uncensored data
        """
        if censored_X.shape[0] == 0:
            self.logger.critical("Nothing to impute, return None")
            return None

        # first learn model without censored data
        self.model.train(uncensored_X, uncensored_y)

        self.logger.debug("Going to impute %d y-values with %s" %
                          (censored_X.shape[0], str(self.model)))

        imputed_y = None  # define this, if imputation fails

        # Define variables
        y = None

        it = 0
        change = 0
        while True:
            self.logger.debug("Iteration %d of %d" % (it, self.max_iter))

            # predict censored y values
            y_mean, y_var = self.model.predict(censored_X)
            y_var[y_var < self.var_threshold] = self.var_threshold
            y_stdev = numpy.sqrt(y_var)

            imputed_y = \
                [truncnorm.stats(a=(censored_y[index] -
                                    y_mean[index]) / y_stdev[index],
                                 b=(self.cutoff * 10 - y_mean[index]) /
                                 y_stdev[index],
                                 loc=y_mean[index],
                                 scale=y_stdev[index],
                                 moments='m')
                 for index in range(len(censored_y))]
            imputed_y = numpy.array(imputed_y)

            if sum(numpy.isfinite(imputed_y) == False) > 0:
                # Replace all nans with threshold
                self.logger.debug("Going to replace %d nan-value(s) with "
                                  "threshold" %
                                  sum(numpy.isfinite(imputed_y) == False))
                imputed_y[numpy.isfinite(imputed_y) == False] = self.threshold

            if it > 0:
                # Calc mean difference between imputed values this and last
                # iteration, assume imputed values are always concatenated
                # after uncensored values

                change = numpy.mean(abs(imputed_y -
                                        y[uncensored_y.shape[0]:]) /
                                    y[uncensored_y.shape[0]:])

            # lower all values that are higher than threshold
            imputed_y[imputed_y >= self.threshold] = self.threshold

            self.logger.debug("Change: %f" % change)

            X = numpy.concatenate((uncensored_X, censored_X))
            y = numpy.concatenate((uncensored_y, imputed_y))

            if change > self.change_threshold or it == 0:
                self.model.train(X, y)
            else:
                break

            it += 1
            if it > self.max_iter:
                break

        self.logger.debug("Imputation used %d/%d iterations, last_change=%f" %
                          (it, self.max_iter, change))

        imputed_y = numpy.array(imputed_y, dtype=numpy.float)
        imputed_y[imputed_y >= self.threshold] = self.threshold

        if not numpy.isfinite(imputed_y).all():
            self.logger.critical("Imputed values are not finite, %s" %
                                 str(imputed_y))
        return imputed_y



import logging
import numpy as np

import smac.runhistory.runhistory
import smac.configspace

__author__ = "Katharina Eggensperger"
__copyright__ = "Copyright 2015, ML4AAD"
__license__ = "AGPLv3"
__maintainer__ = "Katharina Eggensperger"
__email__ = "eggenspk@cs.uni-freiburg.de"
__version__ = "0.0.1"


class RandomEpm(object):
    '''implement an epm, which returns only random values'''

    def __init__(self, seed):
        '''
        initialize random number generator and logger

        Parameters
        ----------
        seed : int
        '''
        self.logger = logging.getLogger("random_epm")
        self.rng = np.random.RandomState(seed)

    def train(self, X, Y, **kwargs):
        '''
        Trains the random forest on X and Y.

        Parameters
        ----------
        X: np.ndarray (N, D)
            Input data points. The dimensionality of X is (N, D),
            with N as the number of points and D is the number of features.
        Y: np.ndarray (N, 1)
            The corresponding target values.
        '''

        if not isinstance(X, np.ndarray):
            raise NotImplementedError("X has to be of type np.ndarray")
        if not isinstance(Y, np.ndarray):
            raise NotImplementedError("Y has to be of type np.ndarray")

        self.logger.debug("Fit model to data")

    def predict(self, X):
        '''
        Predict values for configs

        Parameters
        ----------
        configs : list
            list of configurations

        instance_features : list
            list of instance features

        Returns
        -------
        predictions
        '''
        if not isinstance(X, np.ndarray):
            raise NotImplementedError("X has to be of type np.ndarray")
        return self.rng.rand(len(X), 1), self.rng.rand(len(X), 1)


# encoding=utf8
import abc
import logging
from scipy.stats import norm
import numpy as np

__author__ = "Aaron Klein"
__copyright__ = "Copyright 2015, ML4AAD"
__license__ = "AGPLv3"
__maintainer__ = "Aaron Klein"
__email__ = "kleinaa@cs.uni-freiburg.de"
__version__ = "0.0.1"


class AbstractAcquisitionFunction(object):
    __metaclass__ = abc.ABCMeta
    long_name = ""

    def __str__(self):
        return type(self).__name__ + " (" + self.long_name + ")"

    def __init__(self, model, **kwargs):
        """
        A base class for acquisition functions.

        Parameters
        ----------
        model : Model object
            Models the objective function.

        """
        self.model = model

        self.logger = logging.getLogger("AcquisitionFunction")

    def update(self, **kwargs):
        """Update the acquisition functions values.

        This method will be called if the model is updated. E.g.
        Entropy search uses it to update it's approximation of P(x=x_min),
        EI uses it to update the current fmin.

        The default implementation takes all keyword arguments and sets the
        respective attributes for the acquisition function object.

        Parameters
        ----------
        kwargs
        """

        for key in kwargs:
            setattr(self, key, kwargs[key])

    def __call__(self, X, derivative=False):
        """
        Computes the acquisition value for a given point X

        Parameters
        ----------
        X : np.ndarray
            The input points where the acquisition function
            should be evaluated. The dimensionality of X is (N, D), with N as
            the number of points to evaluate at and D is the number of
            dimensions of one X.

        derivative : Boolean
            If is set to true also the derivative of the acquisition
            function at X is returned
        """

        if len(X.shape) == 1:
            X = X[np.newaxis, :]

        acq = self._compute(X, derivative)
        if np.any(np.isnan(acq)):
            idx = np.where(np.isnan(acq))[0]
            acq[idx, :] = -np.finfo(np.float).max
        return acq

    @abc.abstractmethod
    def _compute(self, X, derivative=False):
        """
        Computes the acquisition value for a given point X. This function has
        to be overwritten in a derived class.

        Parameters
        ----------
        X : np.ndarray
            The input points where the acquisition function
            should be evaluated. The dimensionality of X is (N, D), with N as
            the number of points to evaluate at and D is the number of
            dimensions of one X.

        derivative : Boolean
            If is set to true also the derivative of the acquisition
            function at X is returned

        Returns
        -------
        np.ndarray :
        """
        raise NotImplementedError()


class EI(AbstractAcquisitionFunction):

    def __init__(self,
                 model,
                 par=0.01,
                 **kwargs):
        r"""
        Computes for a given x the expected improvement as
        acquisition value.
        :math:`EI(X) :=
            \mathbb{E}\left[ \max\{0, f(\mathbf{X^+}) -
                f_{t+1}(\mathbf{X}) - \xi\right] \} ]`, with
        :math:`f(X^+)` as the incumbent.

        Parameters
        ----------
        model: Model object
            A model that implements at least
                 - predict(X)
                 - getCurrentBestX().
            If you want to calculate derivatives than it should also support
                 - predictive_gradients(X)
        par: float
            Controls the balance between exploration
            and exploitation of the acquisition function. Default is 0.01
        """

        super(EI, self).__init__(model)
        self.long_name = 'Expected Improvement'
        self.par = par
        self.eta = None

    def _compute(self, X, derivative=False, **kwargs):
        """
        Computes the EI value and its derivatives.

        Parameters
        ----------
        X: np.ndarray(N, D), The input points where the acquisition function
            should be evaluated. The dimensionality of X is (N, D), with N as
            the number of points to evaluate at and D is the number of
            dimensions of one X.

        derivative: Boolean
            If is set to true also the derivative of the acquisition
            function at X is returned

        Returns
        -------
        np.ndarray(N,1)
            Expected Improvement of X
        np.ndarray(N,1)
            Derivative of Expected Improvement at X (only if derivative=True)
        """

        if len(X.shape) == 1:
            X = X[:, np.newaxis]

        m, v = self.model.predict_marginalized_over_instances(X)
        assert m.shape[1] == 1
        assert v.shape[1] == 1
        s = np.sqrt(v)

        if self.eta is None:
            raise ValueError('No current best specified. Call update('
                             'eta=<int>) to inform the acquisition function '
                             'about the current best value.')

        z = (self.eta - m - self.par) / s
        f = (self.eta - m - self.par) * norm.cdf(z) + s * norm.pdf(z)
        f[s == 0.0] = 0.0

        if derivative:
            dmdx, ds2dx = self.model.predictive_gradients(X)
            dmdx = dmdx[0]
            ds2dx = ds2dx[0][:, None]
            dsdx = ds2dx / (2 * s)
            df = (-dmdx * norm.cdf(z) + (dsdx * norm.pdf(z))).T
            df[s == 0.0] = 0.0

        if (f < 0).any():
            self.logger.error("Expected Improvement is smaller than 0!")
            raise ValueError

        if derivative:
            return f, df
        else:
            return f


class EIPS(EI):
    def __init__(self,
                 model,
                 par=0.01,
                 **kwargs):
        r"""
        Computes for a given x the expected improvement as
        acquisition value.
        :math:`EI(X) :=
            \frac{\mathbb{E}\left[ \max\{0, f(\mathbf{X^+}) -
                  f_{t+1}(\mathbf{X}) - \xi\right] \} ]}
                  {np.log10(r(x))}`,
        with :math:`f(X^+)` as the incumbent and :math:`r(x)` as runtime.

        Parameters
        ----------
        model: Model object
            A model that implements at least
                 - predict(X)
                 - getCurrentBestX().
            If you want to calculate derivatives than it should also support
                 - predictive_gradients(X)
        par: float
            Controls the balance between exploration
            and exploitation of the acquisition function. Default is 0.01
        """

        super(EIPS, self).__init__(model)
        self.long_name = 'Expected Improvement per Second'

    def _compute(self, X, derivative=False, **kwargs):
        """
        Computes the EIPS value.

        Parameters
        ----------
        X: np.ndarray(N, D), The input point where the acquisition function
            should be evaluate. The dimensionality of X is (N, D), with N as
            the number of points to evaluate at and D is the number of
            dimensions of one X.

        derivative: Boolean
            Raises NotImplementedError if True.

        Returns
        -------
        np.ndarray(N,1)
            Expected Improvement per Second of X
        """

        if derivative:
            raise NotImplementedError()

        if len(X.shape) == 1:
            X = X[:, np.newaxis]

        m, v = self.model.predict_marginalized_over_instances(X)
        assert m.shape[1] == 2
        assert v.shape[1] == 2
        m_cost = m[:, 0]
        v_cost = v[:, 0]
        # The model already predicts log(runtime)
        m_runtime = m[:, 1]
        s = np.sqrt(v_cost)

        if self.eta is None:
            raise ValueError('No current best specified. Call update('
                             'eta=<int>) to inform the acquisition function '
                             'about the current best value.')

        z = (self.eta - m_cost - self.par) / s
        f = (self.eta - m_cost - self.par) * norm.cdf(z) + s * norm.pdf(z)
        f = f / m_runtime
        f[s == 0.0] = 0.0

        if (f < 0).any():
            self.logger.error("Expected Improvement per Second is smaller than "
                              "0!")
            raise ValueError

        return f.reshape((-1, 1))


import re
import tempfile
import os


RUNHISTORY_FILEPATTERN = '.runhistory_%d.json'
RUNHISTORY_RE = r'\.runhistory_[0-9]*\.json'


def read(run_history, output_directory, configuration_space, logger):
    """Update runhistory with run results from concurrent runs of pSMAC.

    Parameters
    ----------
    run_history : smac.runhistory.RunHistory
        RunHistory object to be updated with run information from runhistory
        objects stored in the output directory.

    output_directory : str
        SMAC output directory. This function will search the output directory
        for files matching the runhistory regular expression.

    configuration_space : ConfigSpace.ConfigurationSpace
        A ConfigurationSpace object to check if loaded configurations are valid.

    logger : logging.Logger

    """
    numruns_in_runhistory = len(run_history.data)
    initial_numruns_in_runhistory = numruns_in_runhistory

    files_in_output_directory = os.listdir(output_directory)
    for file_in_output_directory in files_in_output_directory:
        match = re.match(RUNHISTORY_RE, file_in_output_directory)
        if match:
            runhistory_file = os.path.join(output_directory,
                                           file_in_output_directory)
            run_history.update_from_json(runhistory_file,
                                         configuration_space)

            new_numruns_in_runhistory = len(run_history.data)
            difference = new_numruns_in_runhistory - numruns_in_runhistory
            logger.debug('Shared model mode: Loaded %d new runs from %s' %
                         (difference, file_in_output_directory))
            numruns_in_runhistory = new_numruns_in_runhistory

    difference = numruns_in_runhistory - initial_numruns_in_runhistory
    logger.debug('Shared model mode: Finished loading new runs, found %d new '
                 'runs.' % difference)


def write(run_history, output_directory, num_run):
    """Write the runhistory to the output directory.

    Overwrites previously outputted runhistories.

    Parameters
    ----------
    run_history : smac.runhistory.RunHistory
        RunHistory object to be saved.

    output_directory : str

    run_run : int
        ID of the current SMAC run.

    """

    output_filename = os.path.join(output_directory,
                                   RUNHISTORY_FILEPATTERN % num_run)

    with tempfile.NamedTemporaryFile('wb', dir=output_directory,
                                     delete=False) as fh:
        temporary_filename = fh.name

    run_history.save_json(temporary_filename)
    os.rename(temporary_filename, output_filename)